@@ -9,11 +9,15 @@ from itertools import combinations
 
 import numpy
 
+from ndd.counter import Counts
 from ndd.data import DataArray
 from ndd.divergence import JSDivergence
-from ndd.estimators import NSB, Plugin, check_estimator
+from ndd.estimators import guess_alphabet_size  # pylint: disable=unused-import
+from ndd.estimators import NSB, AutoEstimator, Plugin, check_estimator
 from ndd.exceptions import EstimatorInputError, PmfError
 
+# from ndd.failing import dump_on_fail
+
 __all__ = [
     'entropy',
     'from_data',
@@ -29,26 +33,69 @@ __all__ = [
 logger = logging.getLogger(__name__)
 
 
-def entropy(pk, k=None, estimator='NSB', return_std=False):
+# @dump_on_fail()
+def entropy(nk, k=None, zk=None, estimator='NSB', return_std=False):
     """
-    Entropy estimate from an array of counts.
+    Bayesian Entropy estimate from an array of counts.
 
-    Return a Bayesian estimate for the entropy of an unknown discrete
-    distribution from an input array of counts pk.
+    The entropy function takes as input a vector of frequency counts `nk`
+    (the observed frequencies for a set of classes or states) and an alphabet
+    size `k` (the number of classes with non-zero probability, including
+    unobserved classes) and returns an entropy estimate (in nats) computed
+    using the Nemenman-Schafee-Bialek (NSB) algorithm:
+
+    >>> import ndd
+    >>> counts = [4, 12, 4, 5, 3, 1, 5, 1, 2, 2, 2, 2, 11, 3, 4, 12, 12, 1, 2]
+    >>> ndd.entropy(counts, k=100)
+    2.8060922529931225
+
+    The uncertainty in the entropy estimate can be quantified using the
+    posterior standad deviation, that is stored as an attribute of the
+    `entropy()` function:
+    >>> entropy.err
+    0.32429359488122139
+
+    If the `k` argument is omitted (the alphabet size can be unknown/infinite)
+    the most appropriate estimator will be selected depending on the sampling
+    regime, and an upper bound estimate for `k` will be used.
+
+    >>> counts = [1]*100 + [2]*10
+    >>> entropy(counts)  # `k` is omitted
+    7.2072993808389789
+    >>> entropy.estimator
+    AsymptoticNSB()
+    >>> entropy.err
+    0.3242935948812214
+
+    When the alphabet size is unknown and no coincidences (bins with
+    counts > 1) can be found in the `counts` array, no entropy estimator can
+    give a reasonable estimate. In this case, the the "plugin" estimate will be
+    returned, with entropy.err set to inf.
+
+    >>> counts = [1]*100
+    >>> entropy(counts)
+    4.605170185988092
+    >>> entropy.estimator
+    Plugin(alpha=None)
+    >>> entropy.err
+    inf
 
     Parameters
     ----------
-    pk : array-like
+    nk : array-like
         The number of occurrences of a set of bins.
     k : int or array-like, optional
         Alphabet size (the number of bins with non-zero probability).
-        Must be >= len(pk). A float is a valid input for whole numbers
-        (e.g. k=1.e3). If an array, set k = numpy.prod(k).
-        Default: k = sum(pk > 0)
-    estimator : str or entropy estimator instance, optional
-        If a string, use the estimator class with the same name and default
-        parameters. Check ndd.entropy_estimators for the available estimators.
-        Default: Nemenman-Shafee-Bialek (NSB) estimator.
+        If an array, set k = numpy.prod(k).
+        Default: use an upper bound estimate for the alphabet size, or
+        switch to the asymptotic NSB estimator in the strongly under-sampled
+        regime.
+    zk : array_like, optional
+        Counts distribution or "multiplicities". If passed, nk contains
+        the observed counts values and len(zk) == len(nk).
+    estimator : str or entropy estimator obj, optional
+        Check `ndd.entropy_estimators` for the available estimators.
+        Default: NSB estimator.
     return_std : boolean, optional
         If True, also return an approximation for the standard deviation
         over the entropy posterior.
@@ -57,26 +104,64 @@ def entropy(pk, k=None, estimator='NSB', return_std=False):
     -------
     entropy : float
         Entropy estimate.
-    std : float, optional
-        Uncertainty in the entropy estimate. Only if `return_std` is True.
+    err : float, optional
+        Bayesian error bound on the entropy estimate.
+        Only if `return_std` is True.
+
+    Notes
+    -----
+    The fit parameters can be retrieved using `entropy` function attributes
+    after a function call:
+
+    entropy.entropy : float
+        The entropy estimate.
+    entropy.err : float
+        Error bound on the entropy value.
+    entropy.estimator : estimator object
+        The entropy estimator object (a sklearn-like Estimator).
+    entropy.k : int or float or None
+        The alphabet size used for the fit.
+    entropy.bounded : bool
+        If the entropy estimate has error bounds. When no coincidences have
+        occurred in the data, the estimate is unbounded.
 
     """
 
+    counts = Counts(nk=nk, zk=zk)
+    nk, zk = counts.multiplicities
+
+    if estimator is None:
+        estimator = 'NSB'
+
+    if (estimator == 'NSB' or isinstance(estimator, NSB)) and k is None:
+        estimator = 'AutoEstimator'
+
     estimator, _ = check_estimator(estimator)
 
-    if k is None:
-        k = sum([1 for x in pk if x > 0])
+    estimator = estimator.fit(nk=nk, zk=zk, k=k)
 
-    estimator = estimator.fit(pk, k=k)
     S, err = estimator.estimate_, estimator.err_
 
-    if numpy.isnan(S):
+    if S is not None and numpy.isnan(S):
         logger.warning('nan value for entropy estimate')
         S = numpy.nan
 
+    if err is not None and numpy.isnan(err):
+        err = numpy.nan
+
+    # annotate the entropy function
+    entropy.entropy = S
+    entropy.std = err
+    entropy.bounded = err is not None and numpy.isfinite(err)
+    if isinstance(estimator, AutoEstimator):
+        entropy.estimator = estimator.estimator
+        entropy.k = estimator.k
+    else:
+        entropy.estimator = estimator
+        entropy.k = k
+
     if return_std:
         if err is not None and numpy.isnan(err):
-            err = numpy.nan
             logger.warning('nan value for entropy posterior std deviation')
         return S, err
 
@@ -123,7 +208,7 @@ def from_data(ar, ks=None, estimator='NSB', axis=0, r=None):
     return estimator(counts, k=k)
 
 
-def jensen_shannon_divergence(pk, k=None, estimator='NSB'):
+def jensen_shannon_divergence(nk, k=None, estimator='NSB'):
     """
     Jensen-Shannon divergence from a m-by-p matrix of counts.
 
@@ -138,7 +223,7 @@ def jensen_shannon_divergence(pk, k=None, estimator='NSB'):
     Parameters
     ----------
 
-    pk : array-like, shape (m, p)
+    nk : array-like, shape (m, p)
         Matrix of frequency counts. Each row corresponds to the number of
         occurrences of a set of bins from a different distribution.
     k : int or array-like, optional
@@ -159,7 +244,7 @@ def jensen_shannon_divergence(pk, k=None, estimator='NSB'):
 
     estimator, _ = check_estimator(estimator)
 
-    estimator = JSDivergence(estimator).fit(pk, k=k)
+    estimator = JSDivergence(estimator).fit(nk, k=k)
     js = estimator.estimate_
 
     if numpy.isnan(js):
@@ -169,15 +254,15 @@ def jensen_shannon_divergence(pk, k=None, estimator='NSB'):
     return js
 
 
-def cross_entropy(pk, qk):
+def cross_entropy(nk, qk):
     """
-    Cross entropy: - sum(pk log(pk/qk))
+    Cross entropy: - sum(nk log(nk/qk))
     Parameters
     ----------
-    pk : array_like
+    nk : array_like
         Probability mass function. Normalize if doesnt sum to 1.
     qk : array_like
-        Probability mass function. Must be len(qk) == len(pk).
+        Probability mass function. Must be len(qk) == len(nk).
 
     Returns
     -------
@@ -186,39 +271,39 @@ def cross_entropy(pk, qk):
 
     """
 
-    if len(qk) != len(pk):
-        raise PmfError('qk and pk must have the same length.')
+    if len(qk) != len(nk):
+        raise PmfError('qk and nk must have the same length.')
 
-    pk = numpy.asarray(pk)
+    nk = numpy.asarray(nk)
     qk = numpy.asarray(qk)
 
-    if any(pk < 0):
-        raise PmfError('pk entries must be positive')
+    if any(nk < 0):
+        raise PmfError('nk entries must be positive')
     if not is_pmf(qk):
         raise PmfError('qk must be a valid PMF (positive, normalized)')
 
-    pk = 1.0 * pk / sum(pk)
+    nk = 1.0 * nk / numpy.sum(nk)
     qk = numpy.log(1.0 * qk)
 
-    return -numpy.sum(pk * qk)
+    return -numpy.sum(nk * qk)
 
 
-def kullback_leibler_divergence(pk, qk, estimator='NSB'):
+def kullback_leibler_divergence(nk, qk, estimator='NSB'):
     """
-    Kullback-Leibler divergence given counts pk and a reference PMF qk.
+    Kullback-Leibler divergence given counts nk and a reference PMF qk.
 
-    Return an estimate of the Kullback-Leibler given an array of counts pk and
+    Return an estimate of the Kullback-Leibler given an array of counts nk and
     a reference PMF qk. The estimate (in nats) is computed as:
-    - S_p - sum(pk * log(qk)) / sum(pk)
-    where S_p is the entropy estimate from counts pk.
+    - S_p - sum(nk * log(qk)) / sum(nk)
+    where S_p is the entropy estimate from counts nk.
 
     Parameters
     ----------
-    pk : array_like
+    nk : array_like
         The number of occurrences of a set of bins.
     qk : array_like
-        Reference PMF in sum(pk log(pk/qk).
-        Must be a valid PMF (non-negative, normalized) and len(qk) = len(pk).
+        Reference PMF in sum(nk log(nk/qk).
+        Must be a valid PMF (non-negative, normalized) and len(qk) = len(nk).
     estimator : str or entropy estimator instance, optional
         If a string, use the estimator class with the same name and default
         parameters. Check ndd.entropy_estimators for the available estimators.
@@ -232,12 +317,12 @@ def kullback_leibler_divergence(pk, qk, estimator='NSB'):
     """
 
     estimator, _ = check_estimator(estimator)
-    pk = numpy.asarray(pk)
+    nk = numpy.asarray(nk)
     k = len(qk)
     if k == 1:  # single bin
         return 0.0
 
-    return cross_entropy(pk, qk) - estimator.fit(pk, k=k).estimate_
+    return cross_entropy(nk, qk) - estimator.fit(nk, k=k).estimate_
 
 
 def interaction_information(ar, ks=None, estimator='NSB', axis=0, r=None):
@@ -535,7 +620,7 @@ def is_pmf(a):
     """If a is a valid probability mass function."""
     a = numpy.float64(a)
     not_negative = numpy.all(a >= 0)
-    normalized = numpy.isclose(sum(a), 1.0)
+    normalized = numpy.isclose(numpy.sum(a), 1.0)
     return not_negative and normalized
 
 
